#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import io
import rpm
import os.path
import logging
import binascii
import textwrap
import subprocess

from diffoscope.tools import tool_required
from diffoscope.tempfiles import get_temporary_directory
from diffoscope.difference import Difference

from .rpm_fallback import AbstractRpmFile
from .utils.archive import Archive

logger = logging.getLogger(__name__)


def convert_header_field(io, header):
    if isinstance(header, list):
        if len(header) == 0:
            io.write("[]")
        else:
            for item in header:
                io.write("\n - ")
                convert_header_field(io, item)
        return

    if isinstance(header, str):
        val = header
    elif isinstance(header, bytes):
        try:
            val = header.decode("utf-8")
        except UnicodeDecodeError:
            val = binascii.hexlify(header).decode("us-ascii")
    else:
        val = repr(header)

    io.write(textwrap.fill(val, 100))


def get_rpm_header(path, ts):
    s = io.StringIO()
    with open(path, "r") as f:
        try:
            hdr = ts.hdrFromFdno(f)
        except rpm.error as e:
            logger.error("reading rpm header failed: %s", str(e))
            return str(e)
        for rpmtag in sorted(rpm.tagnames):
            if rpmtag not in hdr:
                continue
            s.write("{}: ".format(rpm.tagnames[rpmtag]))
            convert_header_field(s, hdr[rpmtag])
            s.write("\n")
    return s.getvalue()


def compare_rpm_headers(path1, path2):
    # compare headers
    with get_temporary_directory(suffix="rpm") as rpmdb_dir:
        rpm.addMacro("_dbpath", rpmdb_dir)
        ts = rpm.TransactionSet()
        ts.setVSFlags(-1)
        header1 = get_rpm_header(path1, ts)
        header2 = get_rpm_header(path2, ts)
    return Difference.from_text(
        header1, header2, path1, path2, source="header"
    )


class RpmContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return ["content"]

    @tool_required("rpm2cpio")
    def extract(self, member_name, dest_dir):
        assert member_name == "content"
        dest_path = os.path.join(dest_dir, "content")
        cmd = ["rpm2cpio", self.source.path]
        with open(dest_path, "wb") as dest:
            subprocess.check_call(cmd, stdout=dest, stderr=subprocess.PIPE)
        return dest_path


class RpmFile(AbstractRpmFile):
    DESCRIPTION = "RPM archives"
    CONTAINER_CLASSES = [RpmContainer]

    def compare_details(self, other, source=None):
        return [compare_rpm_headers(self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2020, 2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import io
import os
import re
import struct
import logging
import subprocess

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.archive import Archive
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)


class CbfsListing(Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._header_re = re.compile(
            r"^.*: ([^,]+, bootblocksize [0-9]+, romsize [0-9]+, offset 0x[0-9A-Fa-f]+)$"
        )

    @tool_required("cbfstool")
    def cmdline(self):
        return ["cbfstool", self.path, "print"]

    def filter(self, line):
        return self._header_re.sub("\\1", line.decode("utf-8")).encode("utf-8")


class CbfsContainer(Archive):
    @tool_required("cbfstool")
    def entries(self, path):
        cmd = ["cbfstool", path, "print", "-k"]
        output = our_check_output(cmd).decode("utf-8")
        header = True
        for line in output.rstrip("\n").split("\n"):
            if header:
                if line.startswith("Name"):
                    header = False
                continue
            name = line.split("\t", 1)[0]
            if name == "(empty)":
                continue
            yield name

    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return list(self.entries(self.source.path))

    @tool_required("cbfstool")
    def extract(self, member_name, dest_dir):
        dest_path = os.path.join(dest_dir, os.path.basename(member_name))
        cmd = [
            "cbfstool",
            self.source.path,
            "extract",
            "-n",
            member_name,
            "-f",
            dest_path,
        ]
        logger.debug("cbfstool extract %s to %s", member_name, dest_path)
        our_check_output(cmd, stderr=subprocess.DEVNULL)
        return dest_path


CBFS_HEADER_MAGIC = 0x4F524243
CBFS_HEADER_VERSION1 = 0x31313131
CBFS_HEADER_VERSION2 = 0x31313132
CBFS_HEADER_SIZE = 8 * 4  # 8 * uint32_t

# On 2015-12-15, the largest image produced by coreboot is 16 MiB
CBFS_MAXIMUM_FILE_SIZE = 24 * 2**20  # 24 MiB


def is_header_valid(buf, size, offset=0):
    (
        magic,
        version,
        romsize,
        bootblocksize,
        align,
        cbfs_offset,
        architecture,
        pad,
    ) = struct.unpack_from("!IIIIIIII", buf, offset)
    return (
        magic == CBFS_HEADER_MAGIC
        and (
            version == CBFS_HEADER_VERSION1 or version == CBFS_HEADER_VERSION2
        )
        and (romsize <= size)
        and (cbfs_offset < romsize)
    )


class CbfsFile(File):
    DESCRIPTION = "Coreboot CBFS filesystem images"
    CONTAINER_CLASSES = [CbfsContainer]

    @classmethod
    def recognizes(cls, file):
        size = os.stat(file.path).st_size
        if size < CBFS_HEADER_SIZE or size > CBFS_MAXIMUM_FILE_SIZE:
            return False
        with open(file.path, "rb") as f:
            # pick at the latest byte as it should contain the relative offset of the header
            f.seek(-4, io.SEEK_END)
            # <pgeorgi> given the hardware we support so far, it looks like
            #           that field is now bound to be little endian
            #   -- #coreboot, 2015-10-14
            rel_offset = struct.unpack("<i", f.read(4))[0]
            if (
                rel_offset < 0
                and -rel_offset > CBFS_HEADER_SIZE
                and -rel_offset < size
            ):
                f.seek(rel_offset, io.SEEK_END)
                logger.debug("looking for header at offset: %x", f.tell())
                if is_header_valid(f.read(CBFS_HEADER_SIZE), size):
                    return True
            elif not file.name.endswith(".rom"):
                return False
            else:
                logger.debug(
                    "CBFS relative offset seems wrong, scanning whole image"
                )
            f.seek(0, io.SEEK_SET)
            offset = 0
            buf = f.read(CBFS_HEADER_SIZE)
            while len(buf) >= CBFS_HEADER_SIZE:
                if is_header_valid(buf, size, offset):
                    return True
                if len(buf) - offset <= CBFS_HEADER_SIZE:
                    buf = f.read(32768)
                    offset = 0
                else:
                    offset += 1
            return False

    def compare_details(self, other, source=None):
        return [Difference.from_operation(CbfsListing, self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import contextlib
import curses
import logging
import shutil
import sys


def line_eraser(fd=sys.stderr) -> bytes:
    eraser = b""  # avoid None to avoid 'NoneType + str/bytes' failures
    if fd.isatty():
        curses.setupterm(fd=fd.fileno())
        eraser = curses.tigetstr("el")

    if not eraser and fd.isatty():
        # is a tty, but doesn't support the proper escape code, so let's fake it

        width = shutil.get_terminal_size().columns
        eraser = b"\r" + (b" " * width) + b"\r"

    return eraser


@contextlib.contextmanager
def setup_logging(debug, log_handler):
    logger = logging.getLogger()
    old_level = logger.getEffectiveLevel()
    logger.setLevel(logging.DEBUG if debug else logging.WARNING)

    ch = log_handler or logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)

    formatter = logging.Formatter(
        line_eraser().decode("ascii")
        + "%(asctime)s %(levelname).1s: %(name)s: %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )
    ch.setFormatter(formatter)
    try:
        yield logger
    finally:
        # restore old logging settings. this helps pytest not spew out errors
        # like "ValueError: I/O operation on closed file", see
        # https://github.com/pytest-dev/pytest/issues/14#issuecomment-272243656
        logger.removeHandler(ch)
        logger.setLevel(old_level)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2021 Chris Lamb <lamby@debian.org>
# Copyright © 2018 Mattia Rizzolo <mattia@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import sys
import shutil
import logging
import tempfile

from .utils import format_bytes

_BASEDIR = None
_FILES = []

logger = logging.getLogger(__name__)


def get_named_temporary_file(*args, **kwargs):
    kwargs["dir"] = kwargs.pop("dir", _get_base_temporary_directory())

    if not kwargs.get("suffix", "_").startswith("_"):
        kwargs["suffix"] = "_{}".format(kwargs["suffix"])

    f = tempfile.NamedTemporaryFile(*args, **kwargs)
    _FILES.append(f.name)

    return f


def get_temporary_directory(*args, **kwargs):
    """
    Create and return a TemporaryDirectory

    Preferably use a "with" statement to control its lifetime:
        with get_temporary_directory() as tmpdir:

    The temporary directory is cleaned up at the end of the "with"
    statement. Otherwise it's cleaned up when the object is garbage collected.

    WARNING, don't do this:

        tmpdir = get_temporary_directory().name

    ... as this may result in the removal of the directory at some unexpected
    time in the future. (Or not at all, depending on whether there are other,
    unexpected, references to this instance.)
    """
    kwargs["dir"] = kwargs.pop("dir", _get_base_temporary_directory())

    if not kwargs.get("suffix", "_").startswith("_"):
        kwargs["suffix"] = "_{}".format(kwargs["suffix"])

    d = tempfile.TemporaryDirectory(*args, **kwargs)

    return d


def clean_all_temp_files():
    logger.debug("Cleaning %d temp files", len(_FILES))

    for x in _FILES:
        try:
            os.unlink(x)
        except FileNotFoundError:
            pass
        except:  # noqa
            logger.exception("Unable to delete %s", x)
    _FILES.clear()

    if _BASEDIR is not None:
        logger.debug(
            "Cleaning top-level temporary directory %s", _BASEDIR.name
        )

        shutil.rmtree(_BASEDIR.name, ignore_errors=True)

    if sys.version_info < (3, 8):
        # Some change in tempfile or weakref handling happened after Python
        # 3.7, resulting in us trying to delete our temporary files multiple
        # times. Unfortunately, the cleanup for TemporaryDirectory does not set
        # "ignore_errors", so we do that here.
        import tempfile

        def _rmtree(*args, **kwargs):
            kwargs.setdefault("ignore_errors", True)
            return shutil.rmtree(*args, **kwargs)

        tempfile._rmtree = _rmtree


def _get_base_temporary_directory():
    global _BASEDIR

    if _BASEDIR is None or not os.path.exists(_BASEDIR.name):
        try:
            # Try and generate a potentially-useful suffix to our temporary directory
            filtered_argv = [x for x in sys.argv if not x.startswith("-")]

            suffix = "_{}".format(
                re.sub(
                    r"[^\w]",
                    "",
                    os.path.basename(os.path.dirname(filtered_argv[-1])),
                )[-10:]
            )
        except IndexError:
            suffix = ""

        # Alias the TemporaryDirectory instance (not the .name instance) as the
        # directory may be reference-counted away.
        _BASEDIR = tempfile.TemporaryDirectory(
            dir=tempfile.gettempdir(),
            prefix="diffoscope_",
            suffix=suffix,
        )
        logger.debug(
            "Created top-level temporary directory: %s (free space: %s)",
            _BASEDIR.name,
            format_bytes(get_tempdir_free_space()),
        )

        with open(os.path.join(_BASEDIR.name, "argv"), "w") as f:
            print("\n".join(sys.argv), file=f)

    return _BASEDIR.name


def get_tempdir_free_space():
    """
    unsigned long f_frsize   Fundamental file system block size.
    fsblkcnt_t    f_blocks   Total number of blocks on file system in units of f_frsize.
    fsblkcnt_t    f_bfree    Total number of free blocks.
    fsblkcnt_t    f_bavail   Number of free blocks available to
                             non-privileged process.
    """
    statvfs = os.statvfs(tempfile.gettempdir())

    return statvfs.f_frsize * statvfs.f_bavail



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Juliana Rodrigues <juliana.orod@gmail.com>
# Copyright © 2017-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re

from xml.parsers.expat import ExpatError

from diffoscope.comparators.utils.file import File
from diffoscope.difference import Difference
from diffoscope.tools import python_module_missing

from .missing_file import MissingFile

try:
    defusedxml = True
    from defusedxml import minidom
except ImportError:
    defusedxml = False
    python_module_missing("defusedxml")
    from xml.dom import minidom


def _format(node):
    """
    Removes *inplace* spaces from minidom.Document

    Args:
        node -- A xml.dom.minidom.Document object

    Returns:
        void
    """
    for n in node.childNodes:
        if n.nodeType == n.TEXT_NODE:
            if n.nodeValue:
                n.nodeValue = n.nodeValue.strip()
        elif n.nodeType == n.ELEMENT_NODE:
            _format(n)


def _parse(file):
    """
    Formats a minidom.Document file and returns XML as string.

    Args:
        file -- An io.TextIOWrapper object

    Returns:
        str: formated string object
    """

    if defusedxml:
        xml = minidom.parse(file, forbid_entities=False, forbid_external=False)
    else:
        xml = minidom.parse(file)

    _format(xml)
    xml.normalize()

    return xml.toprettyxml(indent=2 * " ", encoding="utf-8").decode("utf-8")


class XMLFile(File):
    """
    XML Files Comparison class

    Attributes:
        FILE_EXTENSION_SUFFIX (str): xml file extension suffix
    """

    DESCRIPTION = "XML files"
    FILE_TYPE_RE = re.compile(r"^(XML \S+ document|GnuCash file)")

    @classmethod
    def recognizes(cls, file):
        """
        Identifies if a given file has XML extension

        Args:
            file - a diffoscope.comparators.utils.file.File object

        Returns:
            False if file is not a XML File, True otherwise
        """

        # Emulate FALLBACK_FILE_EXTENSION_SUFFIX = {".xml"}
        if not super().recognizes(file) and not file.name.endswith(".xml"):
            return False

        with open(file.path) as f:
            try:
                file.parsed = _parse(f)
            except (ExpatError, UnicodeDecodeError):
                return False

        return True

    def compare_details(self, other, source=None):
        """
        Compares self.object with another, returning a Difference object

        Args:
            other    -- A XMLFile object
            source

        Returns:
            A diffoscope.difference.Difference object
        """
        if isinstance(other, MissingFile):
            return [
                Difference(
                    self.name,
                    other.name,
                    comment="Trying to compare two non-existing files.",
                )
            ]

        difference = Difference.from_text(
            self.dumps(self), self.dumps(other), self.name, other.name
        )

        if difference:
            difference.check_for_ordering_differences()

        return [difference]

    def dumps(self, file):
        """
        Opens a XMLFile and returns its parsed content

        Args:
            file -- XMLFile object

        Returns:
            str -- Formatted XML content from file
        """
        if file.parsed:
            return file.parsed

        with open(file.path) as f:
            return _parse(f)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging
import subprocess

from diffoscope.exc import RequiredToolNotFound
from diffoscope.tools import tool_required
from diffoscope.tempfiles import get_temporary_directory
from diffoscope.difference import Difference

from .utils.archive import Archive
from .utils.file import File
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)


class SevenZList(Command):
    @tool_required("7z")
    def cmdline(self):
        return (
            "7z",
            "l",
            self.path,
        )

    def filter(self, line):
        val = line.decode("utf-8")
        if val.startswith("Listing archive: ") or val.startswith("Path = "):
            return b""
        return line


class SevenZContainer(Archive):
    @tool_required("7z")
    def open_archive(self):
        self._temp_dir = get_temporary_directory(suffix="7z")

        try:
            our_check_output(
                ("7z", "e", os.path.abspath(self.source.path)),
                cwd=self._temp_dir.name,
                stderr=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError:
            return False

        return self

    def close_archive(self):
        self._temp_dir.cleanup()

    def get_member_names(self):
        return os.listdir(self._temp_dir.name)

    def extract(self, member_name, dest_dir):
        return os.path.join(self._temp_dir.name, member_name)


class SevenZFile(File):
    DESCRIPTION = "Filesystem image"
    FILE_TYPE_RE = re.compile(r"^DOS/MBR boot sector;")
    CONTAINER_CLASSES = [SevenZContainer]

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                SevenZList, self.path, other.path, source="7z l"
            )
        ]

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        try:
            cmd = SevenZList(file.path)
            cmd.start()
        except RequiredToolNotFound:
            return False

        return b"Type = gzip\n" not in cmd.output



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2017, 2019-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import logging

from diffoscope.config import Config
from diffoscope.difference import Difference

from .binary import FilesystemFile
from .utils.file import File

logger = logging.getLogger(__name__)


class AbstractMissingType:
    pass


class MissingFile(File, AbstractMissingType):
    """
    Represents a missing file when comparing containers.
    """

    @classmethod
    def recognizes(cls, file):
        if isinstance(file, FilesystemFile) and not os.path.lexists(file.name):
            assert Config().new_file, f"{file.name} does not exist"
            return True
        return False

    def __init__(self, path, other_file=None):
        super().__init__()
        self._name = path
        self._other_file = other_file

    @property
    def path(self):
        return "/dev/null"

    @property
    def other_file(self):
        return self._other_file

    @other_file.setter
    def other_file(self, value):
        self._other_file = value

    def has_same_content_as(self, other):
        return False

    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False

    def is_socket_or_fifo(self):
        return False

    def compare(self, other, source=None):
        # So now that comparators are all object-oriented, we don't have any
        # clue on how to perform a meaningful comparison right here. So we are
        # good do the comparison backward (where knowledge of the file format
        # lies) and and then reverse it.
        if isinstance(other, MissingFile):
            return Difference(
                self.name,
                other.name,
                comment="Trying to compare two non-existing files.",
            )

        logger.debug("Performing backward comparison")
        backward_diff = other.compare(self, source)

        if not backward_diff:
            return None

        return backward_diff.get_reverse()

    # Be nice to text comparisons
    @property
    def encoding(self):
        return self._other_file.encoding

    # Be nice to device comparisons
    def get_device(self):
        return ""

    # Be nice to metadata comparisons
    @property
    def magic_file_type(self):
        return self._other_file.magic_file_type



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Jean-Romain Garnier <salsa@jean-romain.com>
# Copyright © 2020-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import abc
import logging

from .utils.file import File
from .utils.operation import Operation
from .utils.container import Container

from diffoscope.difference import Difference
from diffoscope.excludes import operation_excluded
from diffoscope.tools import (
    tool_required,
    tool_check_installed,
    python_module_missing,
)

try:
    import tlsh
except:
    tlsh = None

try:
    import r2pipe
except:
    python_module_missing("r2pipe")
    r2pipe = None


logger = logging.getLogger(__name__)


if not tool_check_installed("radare2"):
    r2pipe = None
    logger.debug("radare2 not found, disabling decompiler")


class Decompile(Operation, metaclass=abc.ABCMeta):
    def __init__(self, file, *args, **kwargs):
        super().__init__(file.path, *args, **kwargs)
        self.file = file

    def start(self):
        logger.debug("Executing %s", self.full_name())
        if not isinstance(self.file, AsmFunction):
            self._stdout = ""
            return

        self._decompile()

    @abc.abstractmethod
    def _decompile(self):
        raise NotImplementedError()

    def should_show_error(self):
        return False

    @property
    def output(self):
        return self._stdout.encode("utf-8").splitlines(True)


class DecompileGhidra(Decompile):
    # Remove addresses from warnings as they can create a lot of irrelevant noise
    _JUMPTABLE_WARNING_RE = re.compile(rb"(^\s*// WARNING:.*)(0x[0-9a-f]+)")

    def _run_r2_command(self):
        self.file.decompiler.jump(self.file.offset)
        output = self.file.decompiler.r2.cmdj("pdgj")

        if not output:
            # Output is None if the pdg command doesn't exist
            output = {
                "errors": [
                    'Missing r2ghidra-dec, install it with "r2pm install r2ghidra-dec"'
                ]
            }

        return output

    @tool_required("radare2")
    def _decompile(self):
        ghidra_output = self._run_r2_command()

        try:
            self._stdout = ghidra_output["code"]
        except KeyError:
            # Show errors on stdout so a failed decompilation for 1 function
            # doesn't stop the diff for the whole file
            self._stdout = "\n".join(ghidra_output["errors"])
            logger.debug(
                "r2ghidra decompiler error for %s: %s",
                self.file.signature,
                self._stdout,
            )

    def name(self):
        return "r2ghidra"

    def full_name(self, *args, **kwargs):
        return "radare2 r2ghidra"

    def filter(self, line):
        return self._JUMPTABLE_WARNING_RE.sub(rb"\g<1>0xX", line)


class DecompileRadare2(Decompile):
    """
    Significantly faster than the ghidra decompiler, but still outputs assembly
    code, with added comments to make it more readable
    """

    def _run_r2_command(self):
        self.file.decompiler.jump(self.file.offset)
        return self.file.decompiler.r2.cmd("pdc")

    @tool_required("radare2")
    def _decompile(self):
        self._stdout = self._run_r2_command()

    def name(self):
        return "disass"

    def full_name(self, *args, **kwargs):
        return "radare2 disass"


class AsmFunction(File):
    DESCRIPTION = "ASM Function"

    # Mapping between the Config().decompiler option and the command class
    DECOMPILE_OPERATIONS = [
        DecompileGhidra,
        DecompileRadare2,
    ]

    def __init__(self, decompiler, data_dict):
        super().__init__(container=decompiler)
        self.data_dict = data_dict
        self.decompiler = decompiler
        self._name = self.func_name

    @property
    def name(self):
        # Multiple functions can have the same name but a different signature,
        # so use the signature as name for diffoscope
        return self.signature

    @property
    def progress_name(self):
        return "{} [{}]".format(
            self.container.source.progress_name, super().progress_name
        )

    @property
    def path(self):
        return self.container.source.path

    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False

    def is_socket_or_fifo(self):
        return False

    if tlsh:

        @property
        def fuzzy_hash(self):
            if not hasattr(self, "_fuzzy_hash"):
                try:
                    hex_digest = tlsh.hash(self.asm.encode())
                except ValueError:
                    # File must contain a certain amount of randomness
                    return None

                # For short files, the hex_digest is an empty string, so turn
                # it into None
                self._fuzzy_hash = hex_digest or None

            return self._fuzzy_hash

    def has_same_content_as(self, other):
        try:
            return self.hex_dump == other.hex_dump
        except AttributeError:
            # 'other' is not a function.
            logger.debug("has_same_content: Not an asm function: %s", other)
            return False

    @classmethod
    def recognizes(cls, file):
        # No file should be recognized as an asm function
        return False

    def compare(self, other, source=None):
        """
        Override file's compare method to get rid of the binary diff fallback,
        as it would be redundant with other outputs
        """
        details = self.compare_details(other, source)
        details = [x for x in details if x]
        if not details:
            return None

        difference = Difference(self.name, other.name, source=source)
        difference.add_details(details)
        return difference

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(x, self, other)
            for x in list(self.DECOMPILE_OPERATIONS)
        ]

    @property
    def func_name(self):
        return self.data_dict["name"]

    @property
    def offset(self):
        return self.data_dict["offset"]

    @property
    def size(self):
        return self.data_dict["size"]

    @property
    def signature(self):
        return self.data_dict["signature"]

    @property
    def hex_dump(self):
        if not hasattr(self, "_hex_dump"):
            self._hex_dump = self.decompiler.dump(self.offset, self.size)

        return self._hex_dump

    @property
    def asm(self):
        if not hasattr(self, "_asm"):
            ops = self.decompiler.disassemble(self.offset)

            self._asm = ""
            for instr in ops:
                try:
                    self._asm += instr["disasm"] + "\n"
                except KeyError:
                    # Invalid instruction
                    self._asm += "invalid\n"

        return self._asm


def all_decompile_operations_are_excluded(file):
    for klass in AsmFunction.DECOMPILE_OPERATIONS:
        name = " ".join(klass(file).full_name())
        if not operation_excluded(name):
            return False

    return True


class DecompilableContainer(Container):
    auto_diff_metadata = False

    # Don't use @tool_required here so subclassing DecompilableContainer
    # doesn't block the new subclass from doing its work if radare2
    # isn't installed
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.debug("Creating DecompileContainer for %s", self.source.path)

        self._functions = {}

        # Skip disassembly (and decompilation) if a dependency is missing
        # or if radare2 commands are excluded
        if r2pipe is None or all_decompile_operations_are_excluded(
            self.source
        ):
            return

        # Use "-2" flag to silence radare2 warnings
        self.r2 = r2pipe.open(self.source.path, flags=["-2"])

        # Run radare2 command which finds the functions in the executable
        self.r2.cmd("aa")  # Analyse all

        # Hide offset in asm as it serves the same purpose as line numbers,
        # which shouldn't be diffed
        self.r2.cmd("e asm.offset = false")

        # In hex dump of function, hide everything but the hex values
        self.r2.cmd(
            "e hex.offset = false;e hex.header = false;e hex.ascii = false"
        )

        # Use radare2 to get the list of functions
        # If there aren't any, cmdj returns None
        functions = self.r2.cmdj("aj") or []
        for f in functions:
            func = AsmFunction(self, f)
            self._functions[func.signature] = func
            logger.debug("Adding function %s", func.signature)

    def cleanup(self):
        self.r2.quit()

    def get_member_names(self):
        return self._functions.keys()

    def get_member(self, member_name):
        return self._functions[member_name]

    def jump(self, offset):
        self.r2.cmd("s {}".format(offset))

    def dump(self, offset, size):
        self.jump(offset)
        return self.r2.cmd("px {}".format(size)).strip()

    def disassemble(self, offset):
        self.jump(offset)
        return self.r2.cmdj("pdfj")["ops"]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import logging

from ..profiling import profile
from ..utils import format_bytes

from .text import TextPresenter
from .json import JSONPresenter
from .html import HTMLPresenter, HTMLDirectoryPresenter
from .markdown import MarkdownTextPresenter
from .restructuredtext import RestructuredTextPresenter

logger = logging.getLogger(__name__)


class PresenterManager:
    _singleton = {}

    def __init__(self):
        self.__dict__ = self._singleton

        if not self._singleton:
            self.reset()

    def reset(self):
        self.config = {}

    def configure(self, parsed_args):
        FORMATS = {
            "text": {
                "klass": TextPresenter,
                "target": parsed_args.text_output,
            },
            "html": {
                "klass": HTMLPresenter,
                "target": parsed_args.html_output,
            },
            "json": {
                "klass": JSONPresenter,
                "target": parsed_args.json_output,
            },
            "markdown": {
                "klass": MarkdownTextPresenter,
                "target": parsed_args.markdown_output,
            },
            "restructuredtext": {
                "klass": RestructuredTextPresenter,
                "target": parsed_args.restructuredtext_output,
            },
            "html_directory": {
                "klass": HTMLDirectoryPresenter,
                "target": parsed_args.html_output_directory,
            },
        }

        self.config = {
            k: v for k, v in FORMATS.items() if v["target"] is not None
        }

        # If no output specified, default to printing --text output to stdout
        if not self.config:
            FORMATS["text"]["target"] = "-"
            self.config["text"] = FORMATS["text"]

        logger.debug(
            "Will generate the following presenter formats: %s",
            ", ".join(self.config.keys()),
        )

    def output(self, difference, parsed_args, has_differences):
        if difference is None:
            return

        for name, data in self.config.items():
            logger.debug("Generating %r output at %r", name, data["target"])

            # As a special case for text format, write an empty file instead of
            # an empty diff (with headers including the path). This lets people
            # test if the file is empty.
            if not has_differences and name == "text":
                target = data["target"]
                if target != "-":
                    open(target, "w").close()
                continue

            with profile("output", name):
                data["klass"].run(data, difference, parsed_args)

            size = "n/a"
            if os.path.isfile(data["target"]):
                size = format_bytes(os.path.getsize(data["target"]))

            logger.debug(
                "Generated %r output at %r (size: %s)",
                name,
                data["target"],
                size,
            )

    def compute_visual_diffs(self):
        """
        Don't waste time computing visual differences if we won't use them.
        """

        return any(
            x["klass"].supports_visual_diffs for x in self.config.values()
        )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2021, 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging
import os.path

from diffoscope.difference import Difference
from diffoscope.tools import python_module_missing
from diffoscope.profiling import profile
from diffoscope.tools import get_comment_for_missing_python_module

from .utils.file import File
from .utils.archive import Archive

try:
    import guestfs
except ImportError:
    python_module_missing("guestfs")
    guestfs = None

logger = logging.getLogger(__name__)


class FsImageContainer(Archive):
    def open_archive(self):
        if not guestfs:
            return None

        self.g = guestfs.GuestFS(python_return_dict=True)
        try:
            # force a check that LIBGUESTFS_CACHEDIR exists. otherwise guestfs
            # will fall back to /var/tmp, which we don't want
            self.g.set_cachedir(os.environ["LIBGUESTFS_CACHEDIR"])
        except KeyError:
            pass
        self.g.add_drive_opts(self.source.path, format="raw", readonly=1)
        try:
            logger.debug("Launching guestfs; this may take some time")
            with profile("command", "guestfs"):
                self.g.launch()
            logger.debug("guestfs successfully launched")
        except RuntimeError:
            logger.exception("guestfs failed to launch")
            logger.error(
                "If memory is too tight for 512 MiB, try running "
                "with LIBGUESTFS_MEMSIZE=256 or lower."
            )
            return None
        devices = self.g.list_devices()
        try:
            self.g.mount_options("ro", devices[0], "/")
        except RuntimeError:
            logger.exception("guestfs count not mount image; invalid file?")
            return None
        self.fs = self.g.list_filesystems()[devices[0]]
        return self

    def close_archive(self):
        if not guestfs:
            return None

        try:
            self.g.umount_all()
            self.g.close()
        except Exception:  # noqa
            pass

    def get_member_names(self):
        if not guestfs:
            return []
        return [os.path.basename(self.source.path) + ".tar"]

    def extract(self, member_name, dest_dir):
        dest_path = os.path.join(dest_dir, member_name)
        logger.debug("filesystem image extracting to %s", dest_path)
        self.g.tar_out("/", dest_path)
        return dest_path


class FsImageFile(File):
    DESCRIPTION = "ext2/ext3/ext4/btrfs/fat filesystems"
    CONTAINER_CLASSES = [FsImageContainer]
    FILE_TYPE_RE = re.compile(
        r"^(Linux.*filesystem data|BTRFS Filesystem|F2FS filesystem).*"
    )

    @classmethod
    def recognizes(cls, file):
        # Avoid DOS / MBR file type as it generate a lot of false positives,
        # manually check "System identifier string" instead
        with open(file.path, "rb") as f:
            f.seek(54)
            if f.read(8) in (b"FAT12   ", b"FAT16   "):
                return True
            f.seek(82)
            if f.read(8) == b"FAT32   ":
                return True
        return super().recognizes(file)

    def compare_details(self, other, source=None):
        differences = []
        my_fs = ""
        other_fs = ""
        if hasattr(self.as_container, "fs"):
            my_fs = self.as_container.fs
        if hasattr(other.as_container, "fs"):
            other_fs = other.as_container.fs
        if my_fs != other_fs:
            differences.append(
                Difference.from_text(
                    my_fs, other_fs, None, None, source="filesystem"
                )
            )
        if not guestfs:
            self.add_comment(get_comment_for_missing_python_module("guestfs"))
        return differences



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016 Ximin Luo <infinity0@debian.org>
# Copyright © 2016, 2018-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command


class LlvmBcAnalyzer(Command):
    @tool_required("llvm-bcanalyzer")
    def cmdline(self):
        return ["llvm-bcanalyzer", "-dump", self.path]

    def filter(self, line):
        if line.decode("utf-8", "ignore").startswith("Summary of "):
            return b"Summary:"
        return line


class LlvmBcDisassembler(Command):
    @tool_required("llvm-dis")
    def cmdline(self):
        # execute llvm-dis from the same directory as the file, so it doesn't
        # embed the whole path, including our tempdir, into the output.
        # this makes it easier to generate reproducible diffs for our tests.
        return [
            "find",
            self.path,
            "-execdir",
            "llvm-dis",
            "-o",
            "-",
            "{}",
            ";",
        ]


class LlvmBitCodeFile(File):
    DESCRIPTION = "LLVM IR bitcode files"
    FILE_TYPE_RE = re.compile(r"^LLVM IR bitcode")

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(LlvmBcAnalyzer, self.path, other.path),
            Difference.from_operation(
                LlvmBcDisassembler, self.path, other.path
            ),
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2023 FC Stegerman <flx@obfusk.net>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os.path
import re
import tempfile
import zipfile

from contextlib import contextmanager

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.command import Command
from .utils.file import File

# The binary representation of this dummy Android XML file:
# <manifest xmlns:android="http://schemas.android.com/apk/res/android" android:versionCode="1" android:versionName="1" android:compileSdkVersion="29" android:compileSdkVersionCodename="10.0.0" package="com.example" platformBuildVersionCode="29" platformBuildVersionName="10.0.0">
#   <uses-sdk android:minSdkVersion="21" android:targetSdkVersion="29"/>
# </manifest>
#
# This is created by running:
#   $ aapt2 link -o output.apk -I /usr/share/android-framework-res/framework-res.apk --manifest AndroidManifest-in.xml
#   $ unzip output.apk AndroidManifest.xml
# The framework-res.apk needed by aapt2 here is provided by the Debian package
# `android-framework-res`.
# See https://salsa.debian.org/reproducible-builds/diffoscope/-/issues/340
AXML = (
    b"\x03\x00\x08\x00\xf0\x03\x00\x00\x01\x00\x1c\x00l\x02\x00\x00"
    b"\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\\\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x1e\x00\x00\x008\x00\x00\x00R\x00\x00\x00"
    b"v\x00\x00\x00\x9c\x00\x00\x00\xd2\x00\x00\x00\xd8\x00\x00\x00"
    b"\xe8\x00\x00\x00\xfa\x00\x00\x00\x14\x01\x00\x00l\x01\x00\x00"
    b"\x80\x01\x00\x00\x92\x01\x00\x00\xc6\x01\x00\x00\xfa\x01\x00\x00\r\x00m\x00"
    b"i\x00n\x00S\x00d\x00k\x00V\x00e\x00r\x00s\x00i\x00o\x00n\x00\x00\x00\x0b\x00"
    b"v\x00e\x00r\x00s\x00i\x00o\x00n\x00C\x00o\x00d\x00e\x00\x00\x00\x0b\x00v\x00"
    b"e\x00r\x00s\x00i\x00o\x00n\x00N\x00a\x00m\x00e\x00\x00\x00\x10\x00t\x00a\x00"
    b"r\x00g\x00e\x00t\x00S\x00d\x00k\x00V\x00e\x00r\x00s\x00i\x00o\x00n\x00"
    b"\x00\x00\x11\x00c\x00o\x00m\x00p\x00i\x00l\x00e\x00S\x00d\x00k\x00V\x00e\x00"
    b"r\x00s\x00i\x00o\x00n\x00\x00\x00\x19\x00c\x00o\x00m\x00p\x00i\x00l\x00e\x00"
    b"S\x00d\x00k\x00V\x00e\x00r\x00s\x00i\x00o\x00n\x00C\x00o\x00d\x00e\x00"
    b"n\x00a\x00m\x00e\x00\x00\x00\x01\x001\x00\x00\x00\x06\x001\x000\x00.\x00"
    b"0\x00.\x000\x00\x00\x00\x07\x00a\x00n\x00d\x00r\x00o\x00i\x00d\x00"
    b"\x00\x00\x0b\x00c\x00o\x00m\x00.\x00e\x00x\x00a\x00m\x00p\x00l\x00"
    b"e\x00\x00\x00*\x00h\x00t\x00t\x00p\x00:\x00/\x00/\x00s\x00c\x00h\x00e\x00"
    b"m\x00a\x00s\x00.\x00a\x00n\x00d\x00r\x00o\x00i\x00d\x00.\x00c\x00o\x00"
    b"m\x00/\x00a\x00p\x00k\x00/\x00r\x00e\x00s\x00/\x00a\x00n\x00d\x00r\x00"
    b"o\x00i\x00d\x00\x00\x00\x08\x00m\x00a\x00n\x00i\x00f\x00e\x00s\x00"
    b"t\x00\x00\x00\x07\x00p\x00a\x00c\x00k\x00a\x00g\x00e\x00\x00\x00\x18\x00"
    b"p\x00l\x00a\x00t\x00f\x00o\x00r\x00m\x00B\x00u\x00i\x00l\x00d\x00V\x00"
    b"e\x00r\x00s\x00i\x00o\x00n\x00C\x00o\x00d\x00e\x00\x00\x00\x18\x00p\x00l\x00"
    b"a\x00t\x00f\x00o\x00r\x00m\x00B\x00u\x00i\x00l\x00d\x00V\x00e\x00r\x00"
    b"s\x00i\x00o\x00n\x00N\x00a\x00m\x00e\x00\x00\x00\x08\x00u\x00s\x00e\x00s\x00"
    b"-\x00s\x00d\x00k\x00\x00\x00\x00\x00\x80\x01\x08\x00 \x00\x00\x00"
    b"\x0c\x02\x01\x01\x1b\x02\x01\x01\x1c\x02\x01\x01p\x02\x01\x01r\x05\x01\x01"
    b"s\x05\x01\x01\x00\x01\x10\x00\x18\x00\x00\x00\x01\x00\x00\x00"
    b"\xff\xff\xff\xff\x08\x00\x00\x00\n\x00\x00\x00\x02\x01\x10\x00"
    b"\xb0\x00\x00\x00\x01\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff"
    b"\x0b\x00\x00\x00\x14\x00\x14\x00\x07\x00\x00\x00\x00\x00\x00\x00"
    b"\n\x00\x00\x00\x01\x00\x00\x00\xff\xff\xff\xff\x08\x00\x00\x10"
    b"\x01\x00\x00\x00\n\x00\x00\x00\x02\x00\x00\x00\x06\x00\x00\x00"
    b"\x08\x00\x00\x03\x06\x00\x00\x00\n\x00\x00\x00\x04\x00\x00\x00"
    b"\xff\xff\xff\xff\x08\x00\x00\x10\x1d\x00\x00\x00\n\x00\x00\x00"
    b"\x05\x00\x00\x00\x07\x00\x00\x00\x08\x00\x00\x03\x07\x00\x00\x00"
    b"\xff\xff\xff\xff\x0c\x00\x00\x00\t\x00\x00\x00\x08\x00\x00\x03\t\x00\x00\x00"
    b"\xff\xff\xff\xff\r\x00\x00\x00\xff\xff\xff\xff\x08\x00\x00\x10"
    b"\x1d\x00\x00\x00\xff\xff\xff\xff\x0e\x00\x00\x00\x07\x00\x00\x00"
    b"\x08\x00\x00\x03\x07\x00\x00\x00\x02\x01\x10\x00L\x00\x00\x00"
    b"\x02\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\x0f\x00\x00\x00"
    b"\x14\x00\x14\x00\x02\x00\x00\x00\x00\x00\x00\x00\n\x00\x00\x00"
    b"\x00\x00\x00\x00\xff\xff\xff\xff\x08\x00\x00\x10\x15\x00\x00\x00"
    b"\n\x00\x00\x00\x03\x00\x00\x00\xff\xff\xff\xff\x08\x00\x00\x10"
    b"\x1d\x00\x00\x00\x03\x01\x10\x00\x18\x00\x00\x00\x02\x00\x00\x00"
    b"\xff\xff\xff\xff\xff\xff\xff\xff\x0f\x00\x00\x00\x03\x01\x10\x00"
    b"\x18\x00\x00\x00\x01\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff"
    b"\x0b\x00\x00\x00\x01\x01\x10\x00\x18\x00\x00\x00\x01\x00\x00\x00"
    b"\xff\xff\xff\xff\x08\x00\x00\x00\n\x00\x00\x00"
)


class ArscDumpWithAapt(Command):
    MASK_STDERR = True

    @tool_required("aapt2")
    def cmdline(self):
        return ["aapt2", "dump", "resources", self.path]


class ArscFile(File):
    DESCRIPTION = "Android package resource table (ARSC)"
    FILE_TYPE_HEADER_PREFIX = b"\x02\x00\x0c\x00"
    FILE_TYPE_RE = re.compile(r"^Android package resource table\b")
    FILE_EXTENSION_SUFFIX = {".arsc"}

    @tool_required("aapt2")
    def compare_details(self, other, source=None):
        with as_apk(self.path) as self_apk, as_apk(other.path) as other_apk:
            return [
                Difference.from_operation(
                    ArscDumpWithAapt, self_apk, other_apk
                )
            ]


@contextmanager
def as_apk(path):
    # Generate a temporary ZIP file containing the resources.arsc file and a
    # dummy AndroidManifest.xml.  The aapt2 dump resources command requires the
    # resources.arsc file to be inside an APK (ZIP file), and for a valid
    # AndroidManifest.xml file to be present as well.  However, the XML file
    # does not actually need to correspond to the resources.arsc file (though
    # this will result in warnings on stderr, which we ignore); providing a
    # dummy XML file produces the exact same dump.
    with open(path, "rb") as fhi:
        with tempfile.TemporaryDirectory() as tdir:
            apk = os.path.join(tdir, "out.apk")
            with zipfile.ZipFile(apk, "w") as zf:
                zf.writestr("AndroidManifest.xml", AXML)
                with zf.open("resources.arsc", "w") as fho:
                    while True:
                        data = fhi.read(4096)
                        if not data:
                            break
                        fho.write(data)
            yield apk



# -*- coding: utf-8 -*-
#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020-2021 Conrad Ratschan <ratschance@gmail.com>
# Copyright © 2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging
import subprocess

from diffoscope.tools import tool_required, tool_check_installed
from diffoscope.difference import Difference

from .utils import command
from .utils.archive import Archive
from .utils.file import File
from .utils.command import Command

logger = logging.getLogger(__name__)


class FitContainer(Archive):
    # Match the image string in dumpimage list output. Example: " Image 0 (ramdisk@0)"
    IMAGE_RE = re.compile(r"^\sImage ([0-9]+) \((.+)\)", re.MULTILINE)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._members = {}

    def open_archive(self):
        return self

    def close_archive(self):
        pass

    @tool_required("dumpimage")
    def get_member_names(self):
        image_info = command.our_check_output(
            ["dumpimage", "-l", self.source.path],
        )
        members = []
        for match in re.finditer(
            self.IMAGE_RE, image_info.decode(encoding="utf-8")
        ):
            pos, member_name = match.group(1, 2)
            # Save mapping of name -> position as dumpimage takes position as an argument
            self._members[member_name] = pos
            members.append(member_name)

        return members

    @tool_required("dumpimage")
    def extract(self, member_name, dest_dir):
        pos = self._members[member_name]
        dest_path = os.path.join(dest_dir, os.path.basename(member_name))
        logger.debug("fit image extracting %s to %s", member_name, dest_path)

        cmd = (
            "dumpimage",
            "-T",
            "flat_dt",
            "-p",
            pos,
            self.source.path,
            "-o",
            dest_path,
        )

        command.our_check_output(cmd)

        # Cannot rely on dumpimage returning a non-zero exit code on failure.
        if not os.path.exists(dest_path):
            raise subprocess.CalledProcessError(1, cmd)

        return dest_path


class FlattenedImageTreeContents(Command):
    @tool_required("dumpimage")
    def cmdline(self):
        return ["dumpimage", "-l", self.path]


class FlattenedImageTreeFile(File):
    """
    Flattened Image Trees (FIT) are a newer boot image format used by U-Boot. This
    format allows for multiple kernels, root filesystems, device trees, boot
    configurations, and checksums to be packaged into one file. It leverages the
    Flattened Device Tree Blob file format.
    """

    DESCRIPTION = "Flattened Image Tree blob files"
    CONTAINER_CLASSES = [FitContainer]

    @classmethod
    def recognizes(cls, file):
        # Detect file magic manually to workaround incorrect detection of devicetree
        # by libmagic when the devicetree structure section is larger than 1MB
        if file.file_header[:4] != b"\xd0\x0d\xfe\xed":
            return False

        if not tool_check_installed("fdtdump"):
            return False

        # Since the file type is the same as a Device Tree Blob, use fdtget (same
        # package as fdtdump) to differentiate between FIT/DTB
        root_nodes = (
            command.our_check_output(["fdtget", file.path, "-l", "/"])
            .decode(encoding="utf-8")
            .strip()
            .split("\n")
        )
        root_props = (
            command.our_check_output(["fdtget", file.path, "-p", "/"])
            .decode(encoding="utf-8")
            .strip()
            .split("\n")
        )

        # Check for mandatory FIT items used in U-Boot's FIT image verification routine
        return "description" in root_props and "images" in root_nodes

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                FlattenedImageTreeContents, self.path, other.path
            )
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command


class Ffprobe(Command):
    MASK_STDERR = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.flag = False

    @property
    def output(self):
        return self._process.stderr.splitlines(True)

    @tool_required("ffprobe")
    def cmdline(self):
        return ("ffprobe", self.path)

    def filter(self, line):
        if self.flag:
            return line
        if line == b"  Metadata:\n":
            self.flag = True
        return b""


class FfprobeFile(File):
    DESCRIPTION = "Multimedia metadata"
    FILE_TYPE_RE = re.compile(r"^Audio file")

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                Ffprobe, self.path, other.path, source="ffprobe"
            )
        ]
